@@ -9,23 +9,25 @@ from typing import (
     AsyncGenerator,
     Awaitable,
     Callable,
-    DefaultDict,
-    Dict,
-    List,
+    cast,
     Optional,
+    TypeVar,
     Union,
 )
 
 import aiohttp
 from bs4 import BeautifulSoup
+from dataclasses_json.api import DataClassJsonMixin
 
 from .api_objects import (
     Conversation,
     ConversationResponse,
     ConversationsResponse,
+    Error,
     MessageAttachmentCreate,
     MessageCreate,
     Picture,
+    ReactorsResponse,
     RealTimeEventStreamEvent,
     SendMessageResponse,
     URN,
@@ -66,6 +68,33 @@ VERIFY_URL = "https://www.linkedin.com/checkpoint/challenge/verify"
 LINKEDIN_BASE_URL = "https://www.linkedin.com"
 API_BASE_URL = f"{LINKEDIN_BASE_URL}/voyager/api"
 
+T = TypeVar("T", bound=DataClassJsonMixin)
+
+
+async def try_from_json(deserialise_to: T, response: aiohttp.ClientResponse) -> T:
+    if response.status < 200 or 300 <= response.status:
+        try:
+            error = Error.from_json(await response.text())
+        except:
+            raise Exception(
+                f"Deserialising to {deserialise_to} failed because response "
+                f"was {response.status}. Details: {await response.text()}"
+            )
+        raise error
+
+    text = await response.text()
+    try:
+        return deserialise_to.from_json(text)
+    except (json.JSONDecodeError, ValueError) as e:
+        try:
+            error = Error.from_json(text)
+        except:
+            raise Exception(
+                f"Deserialising to {deserialise_to} failed. Error: {e}. "
+                f"Response: {text}."
+            )
+        raise error
+
 
 class ChallengeException(Exception):
     pass
@@ -73,9 +102,16 @@ class ChallengeException(Exception):
 
 class LinkedInMessaging:
     session: aiohttp.ClientSession
-    two_factor_payload: Dict[str, Any]
-    event_listeners: DefaultDict[
-        str, List[Callable[[RealTimeEventStreamEvent], Awaitable[None]]]
+    two_factor_payload: dict[str, Any]
+    event_listeners: defaultdict[
+        str,
+        list[
+            Union[
+                Callable[[RealTimeEventStreamEvent], Awaitable[None]],
+                Callable[[asyncio.exceptions.TimeoutError], Awaitable[None]],
+                Callable[[Exception], Awaitable[None]],
+            ]
+        ],
     ]
 
     def __init__(self):
@@ -228,8 +264,10 @@ class LinkedInMessaging:
             "createdBefore": int(last_activity_before.timestamp() * 1000),
         }
 
-        conversations_resp = await self._get("/messaging/conversations", params=params)
-        return ConversationsResponse.from_json(await conversations_resp.text())
+        res = await self._get("/messaging/conversations", params=params)
+        return cast(
+            ConversationsResponse, await try_from_json(ConversationsResponse, res)
+        )
 
     async def get_all_conversations(self) -> AsyncGenerator[Conversation, None]:
         """
@@ -248,7 +286,10 @@ class LinkedInMessaging:
             if len(conversations_response.elements) < 20:
                 break
 
-            last_activity_before = conversations_response.elements[-1].last_activity_at
+            if last_activity_at := conversations_response.elements[-1].last_activity_at:
+                last_activity_before = last_activity_at
+            else:
+                break
 
     async def get_conversation(
         self,
@@ -271,11 +312,24 @@ class LinkedInMessaging:
             "createdBefore": int(created_before.timestamp() * 1000),
         }
 
-        conversations_resp = await self._get(
+        res = await self._get(
             f"/messaging/conversations/{conversation_urn.id_parts[0]}/events",
             params=params,
         )
-        return ConversationResponse.from_json(await conversations_resp.text())
+        return cast(
+            ConversationResponse, await try_from_json(ConversationResponse, res)
+        )
+
+    async def mark_conversation_as_read(self, conversation_urn: URN) -> bool:
+        res = await self._post(
+            f"/messaging/conversations/{conversation_urn.id_parts[-1]}",
+            json={"patch": {"$set": {"read": True}}},
+        )
+        return res.status == 200
+
+    # endregion
+
+    # region Messages
 
     async def upload_media(
         self,
@@ -316,13 +370,13 @@ class LinkedInMessaging:
 
     async def send_message(
         self,
-        conversation_urn_or_recipients: Union[URN, List[URN]],
+        conversation_urn_or_recipients: Union[URN, list[URN]],
         message_create: MessageCreate,
     ) -> SendMessageResponse:
         params = {"action": "create"}
         message_create_key = "com.linkedin.voyager.messaging.create.MessageCreate"
 
-        message_event: Dict[str, Any] = {
+        message_event: dict[str, Any] = {
             "eventCreate": {"value": {message_create_key: message_create.to_dict()}}
         }
 
@@ -346,9 +400,19 @@ class LinkedInMessaging:
                 f"/messaging/conversations/{conversation_id}/events",
                 params=params,
                 json=message_event,
+                headers=REQUEST_HEADERS,
             )
 
-        return SendMessageResponse.from_json(await res.text())
+        return cast(SendMessageResponse, await try_from_json(SendMessageResponse, res))
+
+    async def delete_message(self, conversation_urn: URN, message_urn: URN) -> bool:
+        res = await self._post(
+            "/messaging/conversations/{}/events/{}".format(
+                conversation_urn, message_urn.id_parts[-1]
+            ),
+            params={"action": "recall"},
+        )
+        return res.status == 204
 
     async def download_linkedin_media(self, url: str) -> bytes:
         async with self.session.get(url) as media_resp:
@@ -360,13 +424,72 @@ class LinkedInMessaging:
 
     # endregion
 
+    # region Reactions
+
+    async def add_emoji_reaction(
+        self,
+        conversation_urn: URN,
+        message_urn: URN,
+        emoji: str,
+    ) -> bool:
+        res = await self._post(
+            "/messaging/conversations/{}/events/{}".format(
+                conversation_urn, message_urn.id_parts[-1]
+            ),
+            params={"action": "reactWithEmoji"},
+            json={"emoji": emoji},
+        )
+        return res.status == 204
+
+    async def remove_emoji_reaction(
+        self,
+        conversation_urn: URN,
+        message_urn: URN,
+        emoji: str,
+    ) -> bool:
+        res = await self._post(
+            "/messaging/conversations/{}/events/{}".format(
+                conversation_urn, message_urn.id_parts[-1]
+            ),
+            params={"action": "unreactWithEmoji"},
+            json={"emoji": emoji},
+        )
+        return res.status == 204
+
+    async def get_reactors(self, message_urn: URN, emoji: str) -> ReactorsResponse:
+        params = {
+            "decorationId": "com.linkedin.voyager.dash.deco.messaging.FullReactor-8",
+            "emoji": emoji,
+            "messageUrn": f"urn:li:fsd_message:{message_urn.id_parts[-1]}",
+            "q": "messageAndEmoji",
+        }
+        res = await self._get("/voyagerMessagingDashReactors", params=params)
+        return cast(ReactorsResponse, await try_from_json(ReactorsResponse, res))
+
+    # endregion
+
+    # region Typing Notifications
+
+    async def set_typing(self, conversation_urn: URN):
+        await self._post(
+            "/messaging/conversations",
+            params={"action": "typing"},
+            json={"conversationId": conversation_urn.get_id()},
+        )
+
+    # endregion
+
     # region Profiles
 
     async def get_user_profile(self) -> UserProfileResponse:
-        response = await self._get("/me")
-        return UserProfileResponse.from_json(await response.text())
+        res = await self._get("/me")
+        return cast(UserProfileResponse, await try_from_json(UserProfileResponse, res))
 
     async def download_profile_picture(self, picture: Picture) -> bytes:
+        if not picture.vector_image:
+            raise Exception(
+                "Failed downloading media. Invalid Picture object with no vector_image."
+            )
         url = (
             picture.vector_image.root_url
             + picture.vector_image.artifacts[-1].file_identifying_url_path_segment
@@ -385,13 +508,28 @@ class LinkedInMessaging:
     def add_event_listener(
         self,
         payload_key: str,
-        fn: Callable[[RealTimeEventStreamEvent], Awaitable[None]],
+        fn: Union[
+            Callable[[RealTimeEventStreamEvent], Awaitable[None]],
+            Callable[[asyncio.exceptions.TimeoutError], Awaitable[None]],
+            Callable[[Exception], Awaitable[None]],
+        ],
     ):
+        """
+        There are three special event types:
+
+        * ALL_EVENTS - an event fired on every event, and which contains the entirety of
+            the raw event payload
+        * TIMEOUT - an event fired if the event listener connection times out
+        * STREAM_ERROR - an event fired if the event stream errors for any other reason
+        """
         self.event_listeners[payload_key].append(fn)
 
     async def _fire(self, payload_key: str, event: Any):
         for listener in self.event_listeners[payload_key]:
-            await listener(event)
+            try:
+                await listener(event)
+            except Exception:
+                logging.exception(f"Listener {listener} failed to handle {event}")
 
     async def _listen_to_event_stream(self):
         logging.info("Starting event stream listener")
@@ -399,6 +537,10 @@ class LinkedInMessaging:
         async with self.session.get(
             REALTIME_CONNECT_URL,
             headers={"content-type": "text/event-stream", **REQUEST_HEADERS},
+            # The event stream normally stays open for about 3 minutes, but this will
+            # automatically close it more agressively so that we don't get into a weird
+            # state where it's not receiving any data, but simultaneously isn't closed.
+            timeout=120,
         ) as resp:
             if resp.status != 200:
                 raise Exception(f"Failed to connect. Status {resp.status}.")
@@ -412,12 +554,22 @@ class LinkedInMessaging:
                     continue
                 data = json.loads(line.decode("utf-8")[6:])
 
+                # Special handling for ALL_EVENTS handler.
+                if all_events_handlers := self.event_listeners.get("ALL_EVENTS"):
+                    for handler in all_events_handlers:
+                        try:
+                            await handler(data)
+                        except Exception:
+                            logging.exception(
+                                f"Handler {handler} failed to handle {data}"
+                            )
+
                 event_payload = data.get(
                     "com.linkedin.realtimefrontend.DecoratedEvent", {}
                 ).get("payload", {})
 
                 for key in self.event_listeners.keys():
-                    if event_payload.get(key):
+                    if event_payload.get(key) is not None:
                         await self._fire(
                             key, RealTimeEventStreamEvent.from_dict(event_payload)
                         )
@@ -428,10 +580,24 @@ class LinkedInMessaging:
         while True:
             try:
                 await self._listen_to_event_stream()
-            except asyncio.exceptions.TimeoutError:
-                continue
+            except asyncio.exceptions.TimeoutError as te:
+                # Special handling for TIMEOUT handler.
+                if all_events_handlers := self.event_listeners.get("TIMEOUT"):
+                    for handler in all_events_handlers:
+                        try:
+                            await handler(te)
+                        except Exception:
+                            logging.exception(
+                                f"Handler {handler} failed to handle {te}"
+                            )
             except Exception as e:
                 logging.exception(f"Error listening to event stream: {e}")
-                continue
+                # Special handling for STREAM_ERROR handler.
+                if all_events_handlers := self.event_listeners.get("STREAM_ERROR"):
+                    for handler in all_events_handlers:
+                        try:
+                            await handler(e)
+                        except Exception:
+                            logging.exception(f"Handler {handler} failed to handle {e}")
 
     # endregion
